"""
mightypy.ml
=============
"""

from ._linear import (
    LinearRegression,
    LassoRegression,
    RidgeRegression,
    LogisticRegression,
    polynomial_regression,
    trend
)


from ._tree import (
    DecisionTreeClassifier,
    DecisionTreeRegressor
)

from ._ensemble import (
    AdaboostClassifier,
    RandomForestClassifier,
    RandomForestRegressor,
)

from ._utils import (
    moving_window_matrix,
    sigmoid
)


__all__ = [
    "LinearRegression",
    "LassoRegression",
    "RidgeRegression",
    "LogisticRegression",
    "polynomial_regression",
    "trend",
    "DecisionTreeClassifier",
    "DecisionTreeRegressor",
    "AdaboostClassifier",
    "RandomForestClassifier",
    "RandomForestRegressor",
    "moving_window_matrix",
    "sigmoid"
]
