"""
mightypy.make
===============
"""


from ._dataset import (
    spiral_data,
    rotation_matrix_2d,
    sine_wave_from_sample,
    sine_wave_from_timesteps
)


__all__ = [
    "spiral_data",
    "rotation_matrix_2d",
    "sine_wave_from_sample",
    "sine_wave_from_timesteps"
]